import { BooleanInput, coerceBooleanProperty } from '@angular/cdk/coercion';
import { Component, ElementRef, HostBinding, Input, Optional, Self, ViewChild, ViewEncapsulation } from '@angular/core';
import { NgControl, ControlValueAccessor } from '@angular/forms';
import { MatFormFieldControl } from '@angular/material/form-field';
import { MatMenuModule } from '@angular/material/menu';
import { Subject } from 'rxjs';

import { OverlayModule } from '@angular/cdk/overlay';
import { DOWN_ARROW, ENTER, ESCAPE, TAB, UP_ARROW } from '@angular/cdk/keycodes';
import { Cursor, MentionHandleResult } from './cursor';

@Component({
  selector: 'app-comment-box',
  imports: [MatMenuModule, OverlayModule],
  providers: [{ provide: MatFormFieldControl, useExisting: CommentBoxComponent }],
  templateUrl: './comment-box.component.html',
  styleUrl: './comment-box.component.scss',
})
export class CommentBoxComponent implements MatFormFieldControl<string>, ControlValueAccessor {

  constructor(@Optional() @Self() public ngControl: NgControl) {
    // Replace the provider from above with this.
    if (this.ngControl != null) {
      // Setting the value accessor directly (instead of using
      // the providers) to avoid running into a circular import.
      this.ngControl.valueAccessor = this;
    }
  }

  ngAfterViewInit(): void {
    if (this.commentText !== '') {
      this._commentInputRef.nativeElement.innerHTML = this.commentText;
      this.onChange(this.commentText);
      this.stateChanges.next()
    }
  }

  ngOnDestroy() {
    this.stateChanges.complete();
  }

  // ==================================================
  // => Everything related to custom form field control
  // ==================================================
  static nextId = 0;
  @HostBinding() id = `comment-with-mentions-${CommentBoxComponent.nextId++}`;
  @HostBinding('class.floating')

  @Input('aria-describedby') userAriaDescribedBy: string = '';
  @Input()
  get value(): string | null {
    return this.commentText;
  }
  set value(text: string | null) {
    this.commentText = text ?? '';
  }
  @Input()
  get placeholder() {
    return this._placeholder;
  }
  set placeholder(plh) {
    this._placeholder = plh;
    this.stateChanges.next();
  }
  get shouldLabelFloat() {
    return this.focused || !this.empty;
  }
  @Input()
  get required(): boolean {
    return this._required;
  }
  set required(req: BooleanInput) {
    this._required = coerceBooleanProperty(req);
    this.stateChanges.next();
  }
  @Input()
  get disabled(): boolean { return this._disabled; }
  set disabled(value: BooleanInput) {
    this._disabled = coerceBooleanProperty(value);
    this.stateChanges.next();
  }

  commentText: string = '';
  controlType = 'comments-input';
  stateChanges = new Subject<void>();
  private _placeholder: string = '';
  private _required = false;
  private _disabled = false;
  focused = false;
  touched = false

  onChange = (text: string) => { };
  onTouched = () => { };

  writeValue(comment: string): void {
    this.commentText = comment;
    // throw new Error('Method not implemented.');
  }
  registerOnChange(onChange: any): void {
    this.onChange = onChange;
    // throw new Error('Method not implemented.');
  }
  registerOnTouched(onTouched: any): void {
    this.onTouched = onTouched;
    // throw new Error('Method not implemented.');
  }
  setDisabledState?(isDisabled: boolean): void {
    this.disabled = isDisabled;
  }

  onFocusIn(event: FocusEvent) {
    if (!this.focused) {
      this.focused = true;
      this.stateChanges.next();
    }
  }

  onFocusOut(event: FocusEvent) {
    if (!this._commentInputRef.nativeElement.contains(event.relatedTarget as Element)) {
      this.cleanEmptyElements(this._commentInputRef.nativeElement);
      this.commentText = this._commentInputRef.nativeElement.innerHTML;
      this.touched = true;
      this.focused = false;
      this.onTouched();
      this.onChange(this.commentText);
      this.stateChanges.next();
    }
  }

  setDescribedByIds(ids: string[]) {
    const controlElement = this._commentInputRef?.nativeElement;
    controlElement?.setAttribute('aria-describedby', ids.join(' '));
  }

  onContainerClick(event: MouseEvent) {
    this._commentInputRef.nativeElement.focus();
  }

  get empty(): boolean {
    if (this._commentInputRef)
      return this._commentInputRef.nativeElement.textContent === ''
    else
      return true;
  }

  get errorState(): boolean {
    return false;
  }


  // ==================================================
  // => Everything related to custom form field control
  // ==================================================

  @ViewChild('commentInput') _commentInputRef!: ElementRef;
  @ViewChild('menuContent') _menuContent!: ElementRef;

  @Input() mentionListItems: string[] = [];

  selectedDropdownItemIndex = 0;
  menuOffsetX = 0;
  menuOffsetY = 0;
  menuOpenAboveYShift = 0;

  isOpen = false;
  filteredSelectionList: string[] = [];

  handleKeyEvent(event: KeyboardEvent) {
    if (this.isOpen) {
      switch (event.keyCode) {
        case DOWN_ARROW:
          this.activateNextItem();
          event.preventDefault();
          break;
        case UP_ARROW:
          this.activatePreviousElement();
          event.preventDefault();
          break;
        case ESCAPE:
          this.removeFocusedMention(event.target as HTMLInputElement);
          event.preventDefault();
          break;
        case ENTER || TAB:
          this.addMention(event.target as HTMLInputElement);
          this.handleTextChanged(true, 1);
          event.preventDefault();
          break;
      }
    }
  }

  clickOutsideMenu(event: any) {
    if (this.isOpen)
      this.removeFocusedMention(this._commentInputRef.nativeElement as HTMLInputElement);
    this.isOpen = false;
  }

  clickMention(selectedIndex: number) {
    this.selectedDropdownItemIndex = selectedIndex;
    this.addMention(this._commentInputRef.nativeElement as HTMLInputElement);
    this.handleTextChanged(true, 1);
  }


  handleInputEvent(event: Event) {
    event.preventDefault();
    this.handleTextChanged();
  }

  handleTextChanged(skipPositionGetting: boolean = false, offset: number = 0) {
    let htmlElement = this._commentInputRef.nativeElement as HTMLInputElement;
    if (!skipPositionGetting)
      offset = Cursor.getPosition(htmlElement);

    // check for new mention Element starters
    let result = this.handleMentionElements(htmlElement.innerHTML);
    htmlElement.innerHTML = result.innerHtml;

    if (result.newOffset)
      offset = result.newOffset;

    // open/close menu
    let focusedMentionElement = <HTMLElement>htmlElement.querySelector(`.mention.${Cursor.focusedClassName}`);
    console.log(focusedMentionElement);
    if (focusedMentionElement) {
      // set offset for menu
      this.menuOffsetX = focusedMentionElement.offsetLeft;
      this.menuOffsetY = focusedMentionElement.offsetTop;
      this.menuOpenAboveYShift = focusedMentionElement.scrollHeight;

      // filter menu entries
      let searchQuery = focusedMentionElement.textContent?.substring(1);
      this.filteredSelectionList = this.mentionListItems.filter((element) => element.toLocaleLowerCase().includes(searchQuery?.toLocaleLowerCase() ?? ""))
      if (this.selectedDropdownItemIndex > this.filteredSelectionList.length - 1)
        this.selectedDropdownItemIndex = this.filteredSelectionList.length - 1;

      this.isOpen = true;
    } else {
      this.isOpen = false;
      this.selectedDropdownItemIndex = 0;
    }

    Cursor.setCurrentCursorPosition(offset, htmlElement);
  }

  handleMentionElements(innerHtml: string): MentionHandleResult {
    let newOffset = null;
    innerHtml = innerHtml.replace(/(?<!(<div class[ \t]*=[ \t]*"[^"]*mention[^"]*">|\w))@\w*/g, (match) => {
      newOffset = match.length;
      return `<div class="${Cursor.focusedClassName} ignore mention">${match}</div>`;
    });

    // remove old focus classes if match was found
    if (newOffset) {
      innerHtml = innerHtml.replaceAll(/class="focused"/g, "");
    }

    return { newOffset, innerHtml };
  }

  activateNextItem() {
    // adjust scrollable-menu offset if the next item is out of view
    let listEl: HTMLElement = this._menuContent.nativeElement;
    let activeEl = listEl.getElementsByClassName('active').item(0);
    if (activeEl) {
      let nextItemEl: HTMLElement = <HTMLElement>activeEl.nextSibling;
      if (nextItemEl && nextItemEl.nodeName === "DIV") {
        let nextLiRect: ClientRect = nextItemEl.getBoundingClientRect();
        if (nextLiRect.bottom > listEl.getBoundingClientRect().bottom) {
          listEl.scrollTop = nextItemEl.offsetTop + nextLiRect.height - listEl.clientHeight;
        }
      }
    }

    if (this.selectedDropdownItemIndex < this.filteredSelectionList.length - 1)
      this.selectedDropdownItemIndex += 1;
    else {
      this.selectedDropdownItemIndex = 0;
      this._menuContent.nativeElement.scrollTop = 0;
    }
  }

  activatePreviousElement() {
    // adjust the scrollable-menu offset if the previous item is out of view
    let listEl: HTMLElement = this._menuContent.nativeElement;
    let activeEl = listEl.getElementsByClassName('active').item(0);
    if (activeEl) {
      let prevItemEl: HTMLElement = <HTMLElement>activeEl.previousSibling;
      if (prevItemEl && prevItemEl.nodeName == "DIV") {
        let prevLiRect: ClientRect = prevItemEl.getBoundingClientRect();
        if (prevLiRect.top < listEl.getBoundingClientRect().top) {
          listEl.scrollTop = prevItemEl.offsetTop;
        }
      }
    }

    if (this.selectedDropdownItemIndex > 0)
      this.selectedDropdownItemIndex -= 1;
    else {
      this.selectedDropdownItemIndex = this.filteredSelectionList.length - 1;
      this._menuContent.nativeElement.scrollTop = listEl.scrollHeight;
    }
  }

  addMention(element: HTMLElement) {
    let focusedMention = element.querySelector(`.mention.${Cursor.focusedClassName}`);
    let focusedElements = element.querySelectorAll(Cursor.focusedClassName);
    if (focusedMention) {
      // set text and remove ignore to style mention
      focusedMention.textContent = `@${this.filteredSelectionList[this.selectedDropdownItemIndex]}`;
      focusedMention.classList.remove(Cursor.focusedClassName);
      focusedMention.classList.remove("ignore")

      // remove all old focused classes
      for (let node of focusedElements) {
        node.classList.remove(Cursor.focusedClassName);
      }
      // create new span next to mention to move cursor there
      let element = document.createElement("span");
      element.classList.add(Cursor.focusedClassName);
      element.innerHTML = "&nbsp;";
      focusedMention.insertAdjacentElement("afterend", element);
    }
  }

  removeFocusedMention(element: HTMLElement) {
    let focused = element.querySelector(`.mention.${Cursor.focusedClassName}`);
    if (focused)
      focused.remove();
  }

  cleanEmptyElements(base: HTMLElement) {
    base.querySelectorAll("span, div").forEach(el => el.textContent === "" && el.parentNode?.removeChild(el))
    base.innerHTML = base.innerHTML.replace(/class=""/g, "");
  }
}